#ifndef _DEFINES_H_
#define _DEFINES_H_

#if !defined(SYS_TIME_H) && !defined(_H_SYS_TIME)
#  include <sys/time.h>
#endif

typedef struct timeval vtime;   /* needs #include <sys/tyme.h> */

#ifdef AIX
#  include <sys/select.h>
#endif

#if !defined(STRING_H) && !defined(_H_STRING)
#  ifdef NO_BCOPY
#    define bcopy(a,b,c) memmove((b), (a), (c))
#  else /* not NO_BCOPY */
     void bcopy();
#  endif /* NO_BCOPY */
#  ifdef NO_BZERO 
#    define bzero(a,b) memset((a), 0, (b))
#  else /* not NO_BZERO */
     void bzero();
#  endif /* NO_BZERO */
#endif /* not !defined(STRING_H) && !defined(_H_STRING) */

#ifdef USE_RANDOM
#  define get_random random
#  define init_random srandom
#else
#  define get_random lrand48
#  define init_random srand48
#endif

#ifdef __STDC__

# ifndef __P
#  define __P(args)				args
# endif
# define __P0(dummy)				(void)
# define __P1(t1,a1)				(t1 a1)
# define __P2(t1,a1,t2,a2)			(t1 a1, t2 a2)
# define __P3(t1,a1,t2,a2,t3,a3)		(t1 a1, t2 a2, t3 a3)
# define __P4(t1,a1,t2,a2,t3,a3,t4,a4)		(t1 a1, t2 a2, t3 a3, t4 a4)
# define __P5(t1,a1,t2,a2,t3,a3,t4,a4,t5,a5)	(t1 a1, t2 a2, t3 a3, t4 a4, t5 a5)

#else /* ! __STDC__ */

# ifndef __P
#  define __P(args)				()
# endif
# define __P0(dummy)				()
# define __P1(t1,a1)				(a1) t1 a1;
# define __P2(t1,a1,t2,a2)			(a1, a2) t1 a1; t2 a2;
# define __P3(t1,a1,t2,a2,t3,a3)		(a1, a2, a3) t1 a1; t2 a2; t3 a3;
# define __P4(t1,a1,t2,a2,t3,a3,t4,a4)		(a1, a2, a3, a4) t1 a1; t2 a2; t3 a3; t4 a4;
# define __P5(t1,a1,t2,a2,t3,a3,t4,a4,t5,a5)	(a1, a2, a3, a4, a5) t1 a1; t2 a2; t3 a3; t4 a4; t5 a5;

#endif /* __STDC__ */

#ifdef __GNUC__
# define INLINE inline
#else
# define INLINE
#endif

#ifndef NULL
#  define NULL ((void *)0)
#endif

#ifndef LONG_MAX
#  define LONG_MAX ( (long) ((~0uL) >> 1) )
#endif

#ifndef LONG_MIN
#  define LONG_MIN ( (long) (((~0uL) >> 1) + 1) )
#endif

#define uSEC_PER_SEC 1000000L  /* microseconds in a second */
#define mSEC_PER_SEC 1000L     /* milliseconds in a second */
#define uSEC_PER_mSEC 1000L    /* microseconds in a millisecond */

#undef MIN2
#undef MAX2
#undef ABS
#undef SIGN
#define MIN2(a,b) ((a)<(b) ? (a) : (b))
#define MAX2(a,b) ((a)>(b) ? (a) : (b))
#define ABS(a)    ((a)> 0  ? (a) :(-a))
#define SIGN(a)   ((a)> 0  ?  1  : (a) ? -1 : 0)

/* macros to match parentheses */
#define ISRPAREN(c) ((c) == ')' || (c) == ']' || (c) == '}')
#define ISLPAREN(c) ((c) == '(' || (c) == '[' || (c) == '{')
#define LPAREN(c) ((c) == ')' ? '(' : ((c) == ']' ? '[' : '{'))

#define ISODIGIT(c) ((c) >= '0' && (c) <= '7')

#define PRINT_INT if(!status) status=1; tty_printf

#define MAXFDSCAN	32	/* max number of fd to scan in select() */
#define ESC		'\\'	/* special escape char */
#define STRESC		"\\"
#define ESC2		'`'	/* other special escape char */
#define STRESC2		"`"
#define CMDSEP		';'	/* command separator character */
#define SPECIAL_CHARS	"{}();\"=" /* specials chars needing escape */
#define CAPLEN		20	/* max length of a terminal capability */
#define SOCKBUFSIZE	4096	/* socket buffer size for read */
#define BUFSIZE		4096	/* general buffer size */
#define LINELEN		BUFSIZE	/* maximum line length */
#define PARAMLEN	999	/* initial length of strings (&x, $x) */
#define MAXMAPLEN	999	/* maximum length of automapped path (<= PARAMLEN) */
#define MIN_WORD_LEN	3	/* the minimum length for history words */
#define WORDSKEPT	512	/* number of words kept for TAB-completion */
#define MAXHIST		128	/* number of history lines kept */
#define MPI		"~$#E"	/* MUME protocol introducer */
#define NUMPARAM	10	/* number of local unnamed params allowed */
#define NUMVAR		50	/* number of global unnamed variables */
#define NUMTOT		(NUMVAR+NUMPARAM)
#define MAXSUBOPT	256	/* max length of suboption string */
#define MAXARGS		16	/* max number of arguments to editor */
#define FLASHDELAY	500	/* time of parentheses flash in millisecs */
#define KBD_TIMEOUT	100	/* timeout for keyboard read in millisecs;
				   hope it's enough also for very slow lines */

#define MAX_STACK	100	/* maximum number of nested
                                 * action, alias, #for or #while */
#define MAX_LOOP	10000	/* maximum number of iterations in
                                 * for or #while */

#define ACTION_WEAK	0	/* GH: normal junk */
#define ACTION_REGEXP	1	/*     oh-so-mighty regexp */
#define ACTION_TYPES	(ACTION_REGEXP + 1)

/* GH: the redefinable delimeters */
#define DELIM		(delim_list[delim_mode])
#define DELIM_LEN	(delim_len[delim_mode])
#define IS_DELIM(c)	(strchr(DELIM, (c)))

#define DELIM_NORMAL	0	/* GH: normal word delimeters	*/
#define DELIM_PROGRAM	1	/*     ()[]{}.,;"'+/-*%		*/
#define DELIM_CUSTOM	2	/*     user-defined		*/
#define DELIM_MODES	(DELIM_CUSTOM + 1)


/* macros to find cursor position from input buffer position */
#define CURLINE(pos)	(((pos) + col0) / (cols - 1) + line0)
#define CURCOL(pos)	(((pos) + col0) % (cols - 1))

#define ISMARKWILDCARD(c) ((c) == '&' || (c) == '$')

/*
 * Attribute codes: bit 0-4 for foreground color or effects, 5-9 for
 * background color. Color #16 is "none", so 0x0210 is "no attribute".
 */
#define COLORS         16         /* number of colors on HFT terminal */
#define LOWCOLORS      8          /* number of ANSI colors */
#define NO_COLOR       COLORS     /* no color change, use default */
#define BITS_COLOR     5          /* bits used for a color entry
                                     (16==none is a valid color) */
#define BITS_2COLOR    10         /* bits used for 2 color entries */
#define COLOR_MASK     0x1F       /* 5 (BITS_COLOR) bits set to 1, others 0 */

#define ATTR_BOLD      0x01
#define ATTR_BLINK     0x02
#define ATTR_UNDERLINE 0x04
#define ATTR_INVERSE   0x08

/*
 * WARNING: colors and attributes are currently using 14 bits.
 * type used is int and 0xFFFFFFFF ==  -1 is used
 * as 'invalid attribute' so in case ints are 16 bits,
 * there is only 1 bit left unused. In case ints are 32 bits, no problem.
 */

/* constructors / accessors for attribute codes */
#define ATTRCODE(attr, fg, bg) \
                (((attr) << BITS_2COLOR) | ((bg) << BITS_COLOR) | (fg))
#define FOREGROUND(attrcode) ((attrcode) & COLOR_MASK)
#define BACKGROUND(attrcode) (((attrcode) >> BITS_COLOR) & COLOR_MASK)
#define ATTR(attrcode)       ((attrcode) >> BITS_2COLOR)

#define NOATTRCODE ATTRCODE(0, NO_COLOR, NO_COLOR)

/*
 * NCSA telnet 2.2 doesn't reset the color when it receives "esc [ m",
 * so we must know what the normal colors are in order to reset it.
 * These colors can be changed with the #color command.
 */
#ifdef BUG_TELNET
# define DEFAULTFG 7		/* make default white text */
# define DEFAULTBG 4		/* on blue background */
#endif

#define LM_NOECHO 1	/* no local echo */
#define LM_CHAR 2	/* char-by-char mode (no line editing) */

#endif /* _DEFINES_H_ */