/*
 * powwow.h
 * various constants etc
 */

#ifndef _POWWOW_H_
#define _POWWOW_H_

#include "defines.h"
#include "ptr.h"

/* generic linked list node (never actually created) */
typedef struct defnode {
    struct defnode *next;
    char *sortfield;
} defnode;

/* linked list nodes: keep the next pointer first, then string to sort by */
typedef struct aliasnode {
    struct aliasnode *next;
    char *name;
    char *subst;
} aliasnode;

typedef struct actionnode {
    struct actionnode *next;
    char *command, *label;
    int active;
    int type;				/* GH: allow regexp */
    char *pattern;
#ifdef USE_REGEXP
    void *regexp;			/* 0 if type == ACTION_WEAK */
#endif
} actionnode;

typedef actionnode promptnode;

typedef struct marknode {
    struct marknode *next;
    char *pattern;
    int attrcode;
} marknode;

typedef struct keynode {
    struct keynode *next;	/* next key definition in list */
    char *name;			/* key name */
    char *sequence;		/* escape sequence sent by terminal */
    int seqlen;			/* GH: length of esc seq to allow \0 in seq */
    void (*funct)();		/* function called when key pressed */
    char *call_data;		/* data passed to function */
} keynode;

typedef struct delaynode {
    struct delaynode *next;
    char *name;
    char *command;
    vtime when;                 /* structure containing time when */
				/* command must be executed */
} delaynode;

/* Variable struct definitions */

typedef struct varnode {        /* for named variables */
    struct varnode *next;
    char *name;
    int index;
    long num;
    ptr  str;
} varnode;

typedef struct {                /* for unnamed vars */
    long num;
    ptr  str;
} unnamedvar;

typedef struct {		/* stack of local vars */
    unnamedvar p[MAX_STACK][NUMPARAM];
    int curr;
} param_stack;

typedef struct {		/* pointers to all variables */
    long *num;
    ptr  *str;
} vars;

/* GH: for keeping track of color status */
typedef struct {
    char *fg, *bg;
    int uline, bold, standout, inv, blink;
} colorstatus;

/* editing session control */
typedef struct editsess {
  int key;		/* session identifier (-1 if cancelled) */
  char builtin;         /* 1 if viewing built-in text */
  int pid;		/* pid of child */
  char *descr;  	/* short description of what we are editing */
  char *file;		/* name of temporary file */
  time_t ctime; 	/* time when temp file was created (upper bound) */
  long oldsize; 	/* original file size */
  struct editsess *next;
} editsess;

/* shared vars from powwow.c */
extern char identified;
extern int  status;
extern char history_done;
extern int  linemode;		/* line mode with (LM_*) flags */
extern char hostname[], portnumber[];
extern char deffile[], helpfile[];

extern int cols, lines;		/* terminal window size */
extern int line0, col0;		/* origin of input line */

extern varnode *prompt;		/* $prompt is always set */
extern varnode *last_line;	/* $line is always set to
				 * the last line processed */
#define promptstr    (ptrdata(prompt->str))
#define promptlen    (ptrlen(prompt->str))
#define promptzero() (ptrzero(prompt->str))

extern char surely_isprompt;    /* 1 if #prompt set #isprompt */
extern char edbuf[];		/* input line buffer */
extern int edlen;		/* characters in edbuf */
extern int pos;			/* cursor position in edbuf */
extern char edattrbeg[], edattrend[];

extern int need_prompt_redraw;	/* GH: set to true when window size changes */
extern int window_size_changed; /* GH: set to true to send new win size */

extern long received, sent;

extern clock_t start_clock, cpu_clock;

extern aliasnode *aliases;
extern actionnode *actions;
extern promptnode *prompts;
extern marknode *markers;
extern int a_nice;
extern keynode *keydefs;
extern delaynode *delays;
extern delaynode *dead_delays;
extern varnode *named_vars[2];
extern int num_named_vars[2];
extern int max_named_vars;
extern param_stack paramstk;
extern vars *var;
#define VAR (var+NUMVAR)

extern vtime now, start_time, ref_time;

extern char initstr[];
extern FILE *capfile, *recordfile;
extern char prefixstr[];
extern char inserted_next[];
extern char verbatim;

extern char opt_exit;
extern char opt_history;
extern char opt_words;
extern char opt_compact;
extern char opt_debug;
extern char opt_wrap;
extern char echo_ext;
extern char echo_int;
extern char echo_key;
extern char opt_speedwalk;
extern char opt_autoprint;

extern char *hist[];
extern int curline;
extern int pickline;

extern void (*last_edit_cmd)();

extern char *delim_list[];
extern int   delim_len[];
extern char *delim_name[];
extern int   delim_mode;

extern char action_chars[];

/* shared functions from powwow.c */

void set_deffile();
int read_settings();
int save_settings();
void signal_init();
void exitpowwow();
void execute();
void syserr();
void errmsg();
char *skipspace();
void process_remote_input();
void unescape();
void put_marks();
void wrap_print();
void smart_print();
void suspend_powwow();
void send_to_main_host();
void setfile();
void strdel();
void strins();
void showprompt();
void parse_user_input();
char *parse_instruction();
void to_input_line();
void cancel_edit();
void push_params();
void pop_params();
void subst_param();
void printver();
void flashparen();
void put_in_history();
char *split_first_word();
int printstrlen();
int parse_walk();
char *str_dup();
void put_word();
void set_custom_delimeters();
void (*sig_permanent())();
int is_permanent_variable();

#endif /* _POWWOW_H_ */
