/*
 * data.c -- basic data structures and functions to manipulate them
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "powwow.h"
#include "eval.h"
/*
 * create a new, empty ptr.
 * return NULL if max == 0
 */
ptr ptrnew __P1 (int,max)
{
    ptr p;
    if (max <= 0)
	return (ptr)0;
    if (max + sizeofptr < max) { /* overflow! */
	error = NO_MEM_ERROR;
	return (ptr)0;
    }
    if ((p = (ptr)malloc(max + sizeofptr))) {
	p->len = 0;
	p->max = max;
	ptrdata(p)[0] = '\0';
    } else
	error = NO_MEM_ERROR;
    return p;
}

/*
 * create a new ptr giving an initial contents,
 * which gets duplicated.
 * 
 * warning: newmax could be so small that we must truncate!
 */
ptr ptrdup2 __P2 (ptr,src, int,newmax)
{
    ptr p;
    
    if (newmax <= 0 || newmax + sizeofptr < 0)
	return (ptr)0;
    if (!src)
	return ptrnew(newmax);

    if ((p = malloc(newmax + sizeofptr))) {
	p->max = newmax;
	if (newmax > ptrlen(src))
	    p->len = newmax = ptrlen(src);
	else
	    p->len = newmax;
	memcpy(ptrdata(p), ptrdata(src), newmax);
	ptrdata(p)[newmax] = '\0';
    } else
	error = NO_MEM_ERROR;
    return p;
}

ptr ptrdup __P1 (ptr,src)
{
    if (!src)
	return src;
    return ptrdup2(src, ptrlen(src));
}

/* delete (free) a ptr */
void ptrdel __P1 (ptr,p)
{
    if (p)
	free((void *)p);
}

/* clear a ptr */
void ptrzero __P1 (ptr,p)
{
    if (p) {
	p->len = 0;
	ptrdata(p)[0] = '\0';
    }
}

/* truncate a ptr to len chars */
void ptrtrunc __P2 (ptr,p, int,len)
{
    if (p) {
	if (len < 0 || len > ptrlen(p))
	    return;
	ptrdata(p)[p->len = len] = '\0';
    } 
}

/* shrink a ptr by len chars */
void ptrshrink __P2 (ptr,p, int,len)
{
    if (p) {
	if (len < 0 || len > ptrlen(p))
	    return;
	ptrdata(p)[p->len -= len] = '\0';
    }
}

/*
 * concatenate two ptr (ptrcat) or a ptr and a char* (ptrmcat)
 * result may be a _newly_allocated_ ptr if original one
 * is too small or if it is soooo big that we are wasting tons of memory.
 * In both cases, the original one may get deleted (freed)
 * You have been warned! Don't use any statically created ptr for
 * write operations, and you will be fine.
 */
ptr __ptrmcat __P4 (ptr,dst, char *,src, int,len, int,shrink)
{
    int newmax, failmax;
    char mustalloc;
    
    if (!src || len<=0)
	return dst;
    if (len + sizeofptr < 0)  /* overflow! */
	return (ptr)0;
    
    if (!dst) {
	mustalloc = 1;
	failmax = len + sizeofptr;
    } else {
	failmax = ptrlen(dst) + len;
	if (failmax + sizeofptr < 0)  /* overflow! */
	    return (ptr)0;
    
	mustalloc = ptrmax(dst) < ptrlen(dst) + len;
    
	if (shrink && ptrmax(dst) > PARAMLEN
	    && ptrmax(dst)/4 > ptrlen(dst) + len)
	    /* we're wasting space, shrink dst */
	    mustalloc = 1;
    }
    
    if (mustalloc) {
	/* dst must be (re)allocated */
	ptr p;

	/* find a suitable new size */
	newmax = PARAMLEN;
	while (newmax + 1 >= 0 && newmax < failmax)
	    newmax <<= 1;
	if (newmax < 0)  /* overflow! */
	    newmax = failmax;
	
	if ((p = (ptr)realloc((void *)dst, newmax + sizeofptr))) {
	    dst = p;
	    dst->max = newmax;
	} else if ((p = ptrdup2(dst, newmax))) {
	    ptrdel(dst);
	    dst = p;
	} else {
	    error = NO_MEM_ERROR;
	    return p;
	}
    }
    if (ptrdata(dst) + ptrlen(dst) != src)
	memmove(ptrdata(dst) + ptrlen(dst), src, len);
    dst->len += len;
    ptrdata(dst)[ptrlen(dst)] = '\0';
    return dst;
}

ptr ptrmcat __P3 (ptr,dst, char *,src, int,len)
{
    return __ptrmcat(dst, src, len, 1);
}

ptr ptrcat __P2 (ptr,dst, ptr,src)
{
    if (src)
	return __ptrmcat(dst, ptrdata(src), ptrlen(src), 1);
    return dst;
}

/*
 * copy a ptr into another (ptrcpy), or a char* into a ptr (ptrmcpy);
 * same warning as above if dst is too small or way too big.
 */
ptr __ptrmcpy __P4(ptr,dst, char *,src, int,len, int,shrink)
{
    int newmax, failmax = len + sizeofptr;
    char mustalloc;
    
    if (!src || len<=0) {
	if (len>=0)
	    ptrzero(dst);
	return dst;
    }
    if (failmax < 0)  /* overflow! */
	return (ptr)0;
    
    if (!dst) {
	mustalloc = 1;
    } else {
	mustalloc = ptrmax(dst) < len;

	if (shrink && ptrmax(dst) > PARAMLEN && ptrmax(dst)/4 > len)
	    /* we're wasting space, shrink dst */
	    mustalloc = 1;
    }
    
    if (mustalloc) {
	/* dst must be (re)allocated */
	ptr p;

	/* find a suitable new size */
	newmax = PARAMLEN;
	while (newmax >= 0 && newmax < failmax)
	    newmax <<= 1;
	
	if (newmax < 0)  /* overflow! */
	    newmax = failmax;
	
	if ((p = (ptr)realloc((void *)dst, newmax + sizeofptr))) {
	    dst = p;
	    dst->max = newmax;
	} else if ((p = ptrdup2(dst, newmax))) {
	    ptrdel(dst);
	    dst = p;
	} else {
	    error = NO_MEM_ERROR;
	    return p;
	}
    }
    if (ptrdata(dst) != src)
	memmove(ptrdata(dst), src, len);
    dst->len = len;
    ptrdata(dst)[ptrlen(dst)] = '\0';
    return dst;
}

ptr ptrmcpy __P3 (ptr,dst, char *,src, int,len)
{
    return __ptrmcpy(dst, src, len, 1);
}

ptr ptrcpy __P2 (ptr,dst, ptr,src)
{
    if (src)
	return __ptrmcpy(dst, ptrdata(src), ptrlen(src), 1);
    ptrzero(dst);
    return dst;
}

/* enlarge a ptr by len chars */
ptr ptrpad __P2 (ptr,p, int,len)
{
    if (!p || len<=0)
	return p;
    if (len > ptrmax(p) - ptrlen(p)) {
	/* must realloc the ptr */
	len += ptrlen(p);
	if (len < 0)  /* overflow! */
	    return NULL;
	/*
	 * cheat: we use ptrmcpy with src==dst
	 * and do an out-of-boud read of src.
	 * works since dst (==src) gets enlarged
	 * before doing the copy.
	 */
	p = ptrmcpy(p, ptrdata(p), len);
    } else {
	p->len += len;
	ptrdata(p)[ptrlen(p)] = '\0';
    }
    return p;
}

/*
 * compare two ptr (ptrcmp) or a ptr and a char* (ptrmcmp)
 * if one is a truncated copy of the other, the shorter is considered smaller
 */
int ptrmcmp __P3 (ptr,p, char *,q, int,lenq)
{
    int res;
    if (!p || ptrlen(p)<=0) {
	if (!q || lenq<=0)
	    /* both empty */
	    res = 0;
    	else
	    res = -1;
    } else if (!q || lenq<=0)
	res = 1;
    else if ((res = memcmp(ptrdata(p), q, MIN2(ptrlen(p), lenq))))
	;
    else if (ptrlen(p) < lenq)
	res = -1;
    else if (ptrlen(p) > lenq)
	res = 1;
    else
	res = 0;
    return res;
}

int ptrcmp __P2 (ptr,p, ptr,q)
{
    if (q)
	return ptrmcmp(p, ptrdata(q), ptrlen(q));
    else if (p)
	return 1;
    else
	return 0;
}

/*
 * find first occurrence of c in p
 * return NULL if none found.
 */
char *ptrchr __P2 (ptr,p, char,c)
{
    if (p)
	return (char *)memchr(ptrdata(p), c, ptrlen(p));
    return (void*)p; /* shortcut for NULL */
}

/*
 * find last occurrence of c in p
 * return NULL if none found.
 */
char *memrchr __P3 (char *,p, int,lenp, char,c)
{
    char *v, *s = p;

    if (!p || lenp<=0)
	return NULL;
    
    v = s + lenp - 1;
    while (v != s && *v != c) {
	v--;
    }
    if (v != s)
	return v;
    else
	return NULL;
}

char *ptrrchr __P2 (ptr,p, char,c)
{
    if (p)
	return memrchr(ptrdata(p), ptrlen(p), c);
    return (void*)p; /* shortcut for NULL */
}

/*
 * find first occurrence of q in p,
 * return NULL if none found.
 */
char *ptrmem __P3 (ptr,p, char *,q, int,lenq)
{
    if (p) {
	if (q && lenq>0)
	    return (char *)memmem(q, lenq, ptrdata(p), ptrlen(p));
	return ptrdata(p);
    }
    return (void*)p; /* shortcut for NULL */
}

char *ptrptr __P2 (ptr,p, ptr,q)
{
    if (p) {
	if (q)
	    return (char *)memmem(ptrdata(q), ptrlen(q), ptrdata(p), ptrlen(p));
	return ptrdata(p);
    }
    return (void*)p; /* shortcut for NULL */
}


/*
 * Scan p for the first occurrence of one of the characters in q,
 * return NULL if none of them is found.
 */
char *memchrs __P4 (char *,p, int,lenp, char *,q, int,lenq)
{
    char *endp;
    
    if (!q || lenq<=0)
	return p;
    if (!p || lenp<=0)
	return NULL;
    
    endp = p + lenp;
    
    while (p < endp && !memchr(q, *p, lenq))
	p++;
    
    if (p == endp)
	return NULL;
    return p;
}

char *ptrmchrs __P3 (ptr,p, char *,q, int,lenq)
{
    if (p)
	return memchrs(ptrdata(p), ptrlen(p), q, lenq);
    return (void*)p; /* shortcut for NULL */
}

char *ptrchrs __P2 (ptr,p, ptr,q)
{
    if (p) {
	if (q)
	    return memchrs(ptrdata(p), ptrlen(p), ptrdata(q), ptrlen(q));
	return ptrdata(p);
    }
    return (void*)p; /* shortcut for NULL */
}


/*
 * Scan p for the last occurrence of one of the characters in q,
 * return NULL if none of them is found.
 */
char *memrchrs __P4 (char *,p, int,lenp, char *,q, int,lenq)
{
    if (!p || lenp<=0) {
	if (!q || lenq<=0)
	    return p;
	else
	    return NULL;
    }
    
    p += lenp;
    if (!q || lenq<=0)
	return p;
    do {
	lenp--, p--;
    } while (lenp >= 0 && !memchr(q, *p, lenq));
	
    if (lenp < 0)
	return NULL;
    return p;
}

char *ptrmrchrs __P3 (ptr,p, char *,q, int,lenq)
{
    if (p)
	return memrchrs(ptrdata(p), ptrlen(p), q, lenq);
    return (void*)p; /* shortcut for NULL */
}

char *ptrrchrs __P2 (ptr,p, ptr,q)
{
    if (p && q)
	return memrchrs(ptrdata(p), ptrlen(p), ptrdata(q), ptrlen(q));
    return p ? ptrdata(p) + ptrlen(p) : (void*)p; /* shortcut for NULL */
}

/*
 * add escapes (backslashes) to a ptr
 */
ptr ptrescape __P2 (ptr,src, ptr,dst)
{
    int left, len;
    char *str, *p; 
    
    if (!src)
	return src;    
    left = ptrlen(src);
    str = ptrdata(src);
    
    if (!dst)
	dst = ptrnew(left*4); /* worst case */
    else {
	dst = ptrpad(dst, left*4 - ptrlen(dst));
    }
    if (error) { errmsg("malloc"); return NULL; }
    
    p = ptrdata(dst);
    
    while (left) {
	if (*str < ' ' || *str > 127 || *str < 0) {
	    sprintf(p, "\\%03o", (int)(unsigned char) *str++);
	    /* warning: accessing private field ->len */
	    len = strlen(p);
	    dst->len += len, p += len, left -= len;
	} else {
	    if (*str == ESC) {
		while (left && *str == ESC)
		    dst->len++, *p++ = *str++, left--;
		
		if (!left || *str == ESC2)
		    dst->len++, *p++ = ESC2;
		else
		    dst->len++, *p++ = ESC;
		    
	    } else if (strchr(SPECIAL_CHARS, *str))
		dst->len++, *p++ = ESC;
	    
	    if (left)
		dst->len++, *p++ = *str++, left--;
	}
    }
    *p = '\0';
    return dst;
}

/*
 * remove escapes (backslashes) from a ptr
 */
void ptrunescape __P1 (ptr,src)
{
    char *p, *s;
    int left;
    enum { NORMAL, ESCAPE } state = NORMAL;

    if (!src)
	return;
    
    left = ptrlen(src);
    p = s = ptrdata(src);

    while (left) switch (state) {
     case NORMAL:
	while (left && *s != ESC) {
	    *p++ = *s++, left--;
	}
	if (left && *s == ESC) {
	    state = ESCAPE;
	    s++, left--;
	}
	break;
     case ESCAPE:
	while (left && *s == ESC) {
	    *p++ = *s++, left--;
	}
	while (left && *s == ESC2) {
	    *p++ = ESC;
	    s++, left--;
	    state = NORMAL; /* redundant... but who cares */
	}
	if (left >= 3 && ISODIGIT(*s) &&
	    ISODIGIT(*(s + 1)) && ISODIGIT(*(s + 2))) {
	    
	    *p++ = (((*s - '0') << 6) |
		    ((*(s + 1) - '0') << 3) |
		    (*(s + 2) - '0'));
	    
	    s += 3, left -= 3;
	    state = NORMAL;
	} else {
	    *p++ = *s++, left--;
	    state = NORMAL;
	}
	break;
     default:
	break;
    }
    src->len = (int)(p - ptrdata(src));
    ptrdata(src)[ptrlen(src)] = '\0';
}

